// wuipath.c - implement WUIMAN path object
// Copyright (c) 1994 by Ron Burk

#include <string.h>
#include "wuipath.h"


// TWuiPath - Handle moving up and down the database "path"
TWuiPath::TWuiPath(const char *TotalPath)
    {   MEMBERASSERT();
    ASSERTDATAPTR(TotalPath);
    ASSERT(TotalPath[0] == '/');
    AbsolutePath_   = new char[MAXPATH+1];
    RelativePath_   = AbsolutePath_;
    ASSERTDATAPTR(AbsolutePath_);
    ASSERTDATAPTR(RelativePath_);
    StringCopy(AbsolutePath_, TotalPath, MAXPATH);
    }

// Copy constructor
TWuiPath::TWuiPath(const TWuiPath &Other)
    :   AbsolutePath_(0), RelativePath_(0)
    {   MEMBERASSERT();
    *this   = Other;
    }

void    TWuiPath::operator=(const TWuiPath &Other)
    {   MEMBERASSERT();
    if(&Other != this)
        {
        if(AbsolutePath_ != NULL)
            delete[] AbsolutePath_;
        AbsolutePath_   = new char[MAXPATH+1];
        ASSERTDATAPTR(AbsolutePath_);
        StringCopy(AbsolutePath_, Other.AbsolutePath(), MAXPATH);
        RelativePath_   = AbsolutePath_ +
                    (Other.RelativePath() - Other.AbsolutePath());
        }
    }

// destructor - free up path string
TWuiPath::~TWuiPath()
    {   MEMBERASSERT();
    ASSERTDATAPTR(AbsolutePath_);
    delete[] AbsolutePath_;
    }

// Descend() - Move relative path down one node.
int TWuiPath::Descend() const
    {   MEMBERASSERT();
    char    *Child = strchr(RelativePath_, '/');
    if(!Child)   // if no children
        return FALSE;
    // ignore trailers (e.g. "/a/b/c/")
    else if(Child[1] == '\0')
        return FALSE;
    // else there's *some* kind of child name
    else
        {
        (char *&)RelativePath_   = Child+1;
        return TRUE;
        }
    }

// Ascend() - Move relative path up one node.
int TWuiPath::Ascend() const
    {   MEMBERASSERT();
    switch(RelativePath_ - AbsolutePath_)
        {
        case    0:
            return FALSE;
        case    1:
            (char *&)RelativePath_   = AbsolutePath_;
            return TRUE;
        default:
            (char *&)RelativePath_   -= 2;
            while(RelativePath_ > AbsolutePath_)
                if(*RelativePath_ == '/')
                    break;
                else
                    --(char *&)RelativePath_;
            ASSERT(*RelativePath_ == '/');
            ++(char *&)RelativePath_;
            return TRUE;
        }
    }

// CurrentName() - return first name of relative path.
TWuiName    TWuiPath::CurrentName() const
    {   MEMBERASSERT();
    char        Name[TWuiName::MAXNAME+1];
    const char *Rover = strchr(RelativePath(), '/');
    int         Length;
    if(Rover)
        Length  = (int)(Rover-RelativePath());
    else
        Length  = strlen(RelativePath());
    const char *Start = RelativePath();
    // AWuiObject::Get/Set needs this
    if(*Start == '.')
        ++Start;
    strncpy(Name, Start, Length);
    Name[Length]    = '\0';
    return TWuiName(Name);
    }

// NextName() - return next name of relative path.
TWuiName    TWuiPath::NextName() const
    {   MEMBERASSERT();
    if(Descend())
        {
        TWuiName Result = CurrentName();
        Ascend();
        return Result;
        }
    else
        return NULL;
    }

// Push() - Append to the path.
int     TWuiPath::Push(const char *SubPath)
    {   MEMBERASSERT();
    char   *Rover = AbsolutePath_ + strlen(AbsolutePath_);
    if(Rover > AbsolutePath_)
        --Rover;    // point to char before EOS
    if(*Rover != '/')
        strcat(AbsolutePath_, "/");
    strcat(AbsolutePath_, SubPath);
    return TRUE;    // Need to check for overflow!
    }
// Push() - Append to the path (TWuiName version)
int     TWuiPath::Push(const TWuiName SubPath)
    {   MEMBERASSERT();
    char    Name[TWuiName::MAXNAME+1];
    return Push(SubPath.GetName(Name));
    }

// Pop() - Pop the last node off the path.
int TWuiPath::Pop()
    {   MEMBERASSERT();
    // if already at root, can't pop
    if(!strcmp(AbsolutePath_, "/"))
        return FALSE;
    char   *Rover = AbsolutePath_ + strlen(AbsolutePath_);
    if(Rover > AbsolutePath_)
        --Rover;    // point to char before EOS
    if(Rover > AbsolutePath_ && *Rover == '/')
        --Rover;    // skip any bogus trailing '/'
    while(Rover > AbsolutePath_ && *Rover != '/')
        --Rover;
    // if string did not start with "/", this is bad!
    if(Rover < AbsolutePath_)
        {
        DEBUG_WARNING("'%s':Bad path in TWuiPath::Pop()",
                                             AbsolutePath_);
        return FALSE;
        }
    else    // else Rover points at '/'
        {
        // as always, special case for root
        if(Rover == AbsolutePath_)
            ++Rover;
        *Rover = '\0';
        RelativePath_   = AbsolutePath_;
        while(Descend())
            ;
        return TRUE;
        }
    }

// IsLastNode(0) is true if this node is last (attribute doesn't count)
// IsLastNode(1) is true if this is next to last node, etc.
int     TWuiPath::IsLastNode(int Which) const
    {   MEMBERASSERT();
    const char *Rover = RelativePath_;
    for(Rover=RelativePath_; Which >= 0; --Which)
        {
        Rover   = strchr(RelativePath_, '/');
        if(Rover != NULL)
            ++Rover;
        else if(Which == 0)
            return TRUE;
        else
            break;
        }
    return FALSE;
    }

// make a copy of the path, up to the current node
void    TWuiPath::CopyRelativePath(char *Output) const
    {   MEMBERASSERT();
    strcpy(Output, AbsolutePath_);
    const char *Rover = RelativePath_;
    // if current node is root
    if(Rover == AbsolutePath_)  // if current node is root
        ++Rover;
    // else if current node is last node
    else if((Rover=strchr(RelativePath_, '/')) == NULL)
        Rover   = RelativePath_ + strlen(RelativePath_);
    ASSERT(Rover != NULL);
    ASSERT(Rover > AbsolutePath_);
    StringCopy(Output, AbsolutePath_, (Rover-AbsolutePath_)+1);
    }



